#ifndef __time_h_embedded__
#define __time_h_embedded__


struct tm {
  int8_t tm_sec;			/* Seconds.	[0-60] (1 leap second)	*/
  int8_t tm_min;			/* Minutes.	[0-59]			*/
  int8_t tm_hour;			/* Hours.	[0-23]			*/
  int8_t tm_mday;			/* Day.		[1-31]			*/
  int8_t tm_mon;			/* Month.	[1-12]			*/
  int16_t tm_year;			/* Year	- 1900.				*/
  int8_t tm_wday;			/* Day of week.	[0-6]	Sunday = 0	*/
  int16_t tm_yday;			/* Days in year.[1-366]			*/
  int8_t tm_isdst;			/* DST.		[-1/0/1]		*/
};


void timestamp2tm (int32_t ts, struct tm *t) {
	// This code will not handle dates < 1970-01-01, i.e. negative timestamps
	if (ts < 0) ts = 0;
	int8_t leap = 0;
	t->tm_sec = 0;
	t->tm_min = 0;
	t->tm_hour = 0;
	t->tm_mday = 1;
	t->tm_mon  = 1;
	t->tm_year = 1970;	// start Thursday 1970-01-01
	if (ts > (60L*60*24*3) )
		t->tm_wday = ( (ts - (60L*60*24*3) ) / (60L * 60 * 24) % 7);
	else			// First week of 1970
		t->tm_wday = (ts / (60L*60*24) ) + 4;

	t->tm_yday = 0;
	t->tm_isdst = 0;	// Not handled!

	while (ts >= (60L*60*24*(365+leap)) ) {
		ts-= (60L*60*24*(365+leap));
		t->tm_year++;
		if ( ((t->tm_year+1) % 4) == 0  ) { leap = 1; } else { leap = 0; }
		if ( ((t->tm_year+1) % 100) == 0) { leap = 0; }
		if ( ((t->tm_year+1) % 400) == 0) { leap = 1; }
	}
	{
		int8_t month_days[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30 ,31, 30, 31};
		int8_t *p_month = &month_days[0];
		if (leap) month_days[1]++;
		while (ts >= (*p_month * 60L * 60 * 24)) {
			t->tm_mon++;
			t->tm_yday += *p_month;
			ts -= (*p_month++ * 60L * 60 * 24);
		}
	}
	while (ts >= (60L*60*24)) {
		t->tm_mday++;
		ts -= (60L*60*24);
	}
	while (ts >= (60L*60)) {
		t->tm_hour++;
		ts -= (60L*60);
	}
	while (ts >= 60) {
		t->tm_min++;
		ts -= 60;
	}
	t->tm_yday += t->tm_mday;
	t->tm_sec = ts;
	t->tm_year = t->tm_year - 1900;
}

#endif

